/*
 * Host Identity Protocol
 * Copyright (C) 2005-06 the Boeing Company
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *  hip_status.c
 *
 *  Authors: Jeff Ahrenholz <jeffrey.m.ahrenholz@boeing.com>
 * 
 * Status thread
 *
 */

#ifdef __CYGWIN__
#include <windows.h>
#include <w32api/iphlpapi.h>
#endif
#include <stdio.h>		/* printf() */
#ifdef __WIN32__
#include <win32/types.h>
#else
#include <unistd.h>
#include <pthread.h>		/* phread_exit() */
#include <sys/uio.h>		/* iovec */
#endif
#include <string.h>		/* strerror() */
#include <errno.h>		/* errno */
#include <hip/hip_service.h>
#include <hip/hip_types.h>
#include <hip/hip_sadb.h>		/* access to SADB */
#include <hip/hip_status.h>
#include <win32/rtnetlink.h>		/* from <linux/rtnetlink.h> */


/* 
 * Local function declarations
 */
void handle_status_request(int type, char *buff, int *len);
void dump_sadb(char *buff, int *tlv_len);
void dump_dst_entries(char *buff, int *tlv_len);
void dump_lsi_entries(char *buff, int *tlv_len);

/*
 * hip_status()
 *
 *
 */
#ifdef __WIN32__
void hip_status(void *arg)
#else
void *hip_status(void *arg)
#endif
{
	int err, s, len;
#ifdef __CYGWIN__
	int from_len;
#else
	socklen_t from_len;
#endif
	char buff[2048];
	fd_set read_fdset;
	struct timeval timeout;
	struct sockaddr_in addr;
	struct sockaddr_storage ss_from;
	struct sockaddr *from = (struct sockaddr*) &ss_from;
	struct status_tlv req;

	printf("hip_status() thread started...\n");
	if ((s = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
		printf("Status thread socket() error: %s\n", strerror(errno));
		fflush(stdout);
#ifdef __WIN32__
		return;
#else
		return NULL;
#endif
	}

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.sin_port = htons(WIN_STATUS_PORT);

	if (bind(s, (struct sockaddr*)&addr, sizeof(addr)) < 0) {
		printf("Status thread: bind() error - %s\n", strerror(errno));
		fflush(stdout);
#ifdef __WIN32__
		return;
#else
		return NULL;
#endif
	}
	
	while(g_state == 0) {
		FD_ZERO(&read_fdset);
		FD_SET((unsigned)s, &read_fdset);
		timeout.tv_sec = 1;
		timeout.tv_usec = 0;

		if ((err = select(s+1, &read_fdset, NULL,NULL, &timeout) < 0)) {
			if (errno == EINTR)
				continue;
			printf("Status thread: select() error: %s.\n",
				strerror(errno));
		} else if (FD_ISSET(s, &read_fdset)) {
			memset(buff, 0, sizeof(buff));
			from_len = sizeof(struct sockaddr_storage);
			if ((err = recvfrom(s, (char*)&req, sizeof(req), 0, 
						from, &from_len)) < 0) {
				if (err != -EINTR)
				   printf("Status thread: recvfrom error: %s\n",
					strerror(errno));
				continue;
			}
			if (err != sizeof(struct status_tlv))
				continue;
			len = sizeof(buff);
			handle_status_request(ntohs(req.tlv_type), buff, &len);
			len = sendto(s, buff, len, 0, from, from_len);
		}

	}

	printf("hip_status() thread shutdown.\n");
	fflush(stdout);
#ifndef __WIN32__
	pthread_exit((void *) 0);
	return(NULL);
#endif
}

void handle_status_request(int type, char *buff, int *len)
{
	int tlv_len=0;
	struct status_tlv *t = (struct status_tlv*) buff;
	
	switch(type) {
		case STAT_THREADS:
			t->tlv_type = htons(HIP_STATUS_REPLY_STRING);
			t->tlv_len = htons(40);
			t++;
			sprintf((char*)t,
				"Thread info not available yet.\n");
			tlv_len = 40;
			break;
		case STAT_SADB:
			dump_sadb(buff, &tlv_len);
			break;
		case STAT_DST:
			dump_dst_entries(buff, &tlv_len);
			break;
		case STAT_LSI:
			dump_lsi_entries(buff, &tlv_len);
			break;
		case STAT_MIN:
		case STAT_MAX:
		default:
			printf("<sending error (%d)>", type);
			t->tlv_type = htons(HIP_STATUS_REPLY_ERROR);
			t->tlv_len = 0;
			t++;
			break;
	}
	t = (struct status_tlv*) ((char*)t + tlv_len);
	t->tlv_type = htons(HIP_STATUS_REPLY_DONE);
	t->tlv_len = 0;
	t++;			
	*len = (char*)t - buff;
}

extern hip_sadb_entry hip_sadb[SADB_SIZE];

int sockaddr_list_length(sockaddr_list *l)
{
	int count = 0;
	for(; l; l=l->next)
		count++;
	return(count);
}

void dump_sadb(char *buff, int *tlv_len)
{
	hip_sadb_entry *entry;
	struct status_tlv *t = (struct status_tlv*)buff;
	int i, len, n;
	char *p;
	sockaddr_list *l;

	for (i=0; i < SADB_SIZE; i++) {
		for (	entry = &hip_sadb[i]; 
			entry && entry->spi; 
			entry=entry->next	) {
			t->tlv_type = htons(HIP_STATUS_REPLY_SADB);
			t->tlv_len = 0;
			p = (char *)(t + 1);
			len = 0;
			ADD_ITEM(p, entry->spi, len);
			ADD_ITEM(p, entry->direction, len);
			ADD_ITEM(p, entry->hit_magic, len);
			ADD_ITEM(p, entry->mode, len);
			ADD_ITEM(p, entry->lsi, len);
			ADD_ITEM(p, entry->lsi6, len);
			ADD_ITEM(p, entry->a_type, len);
			ADD_ITEM(p, entry->e_type, len);
			ADD_ITEM(p, entry->a_keylen, len);
			ADD_ITEM(p, entry->e_keylen, len);
			ADD_ITEM(p, entry->lifetime, len);
			ADD_ITEM(p, entry->bytes, len);
			ADD_ITEM(p, entry->sequence, len);
			/*ADD_ITEM(p, entry->replay_win, len);
			ADD_ITEM(p, entry->replay_map, len);
			ADD_ITEM(p, entry->iv, len);*/
			n = sockaddr_list_length(entry->src_addrs);
			ADD_ITEM(p, n, len);
			n = sockaddr_list_length(entry->dst_addrs);
			ADD_ITEM(p, n, len);
			t->tlv_len = htons((__u16)len);
			t = (struct status_tlv *)(p + len);
			
			/* addresses */
			t->tlv_type = htons(HIP_STATUS_REPLY_ADDR);
			t->tlv_len = 0;
			p = (char *)(t + 1);
			len = 0;			
			for (l = entry->src_addrs; l; l=l->next)
				ADD_ITEM(p, l->addr, len);
			for (l = entry->dst_addrs; l; l=l->next)
				ADD_ITEM(p, l->addr, len);
			/* TODO: add NAT variables here */
			t->tlv_len = htons((__u16)len);
			t = (struct status_tlv *)(p + len);

		}
		
	}
	*tlv_len = (char*)t - buff;
}

extern hip_sadb_dst_entry hip_sadb_dst[SADB_SIZE];
void dump_dst_entries(char *buff, int *tlv_len)
{
	hip_sadb_dst_entry *entry;
	struct status_tlv *t = (struct status_tlv*)buff;
	int i, len;
	char *p;

	for (i=0; i < SADB_SIZE; i++) {
		for (	entry = &hip_sadb_dst[i]; 
			entry && entry->sadb_entry; 
			entry=entry->next	) {
			t->tlv_type = htons(HIP_STATUS_REPLY_DST_ENTRY);
			t->tlv_len = 0;
			p = (char *)(t + 1);
			len = 0;
			ADD_ITEM(p, entry->addr, len);
			ADD_ITEM(p, entry->sadb_entry->spi, len);
			t->tlv_len = htons((__u16)len);
			t = (struct status_tlv *)(p + len);
		}
	}
	*tlv_len = (char*)t - buff;
}

extern hip_lsi_entry *lsi_temp;
void dump_lsi_entries(char *buff, int *tlv_len)
{
	hip_lsi_entry *l;
	struct status_tlv *t = (struct status_tlv*)buff;
	int len;
	char *p;

	for (l = lsi_temp; l; l=l->next) {
			t->tlv_type = htons(HIP_STATUS_REPLY_LSI_ENTRY);
			t->tlv_len = 0;
			p = (char *)(t + 1);
			len = 0;
			ADD_ITEM(p, l->addr, len);
			ADD_ITEM(p, l->lsi4, len);
			ADD_ITEM(p, l->lsi6, len);
			ADD_ITEM(p, l->num_packets, len);
			ADD_ITEM(p, l->next_packet, len);
			ADD_ITEM(p, l->send_packets, len);
			ADD_ITEM(p, l->creation_time.tv_sec, len);
			t->tlv_len = htons((__u16)len);
			t = (struct status_tlv *)(p + len);
	}

	*tlv_len = (char*)t - buff;
}

